!Skew-T Program
!Designed to read in an ascii text file holding an atmospheric sounding,
!and output an ascii file for easy plotting, such as by Excel.
!Copyright  2007 by Roland Stull
!version 14.  4 Feb 2007.  Modified 5 Aug 2021.


!============ modules =====================

module soundmod							!holds sounding arrays
	character (len=50) :: filename  	!holds name of sounding file
	character (len=60) :: outname  		!holds name of output file
	character (len=100) :: title		!title line for sounding
	integer :: nlines					!number of lines in sounding, initialized to 0
	integer, parameter :: maxlines = 120   !max sounding lines that can be captured
	real, dimension(maxlines) :: P		!Pressure (kPa)
	real, dimension(maxlines) :: z		!Height (m)
	real, dimension(maxlines) :: T		!Temperature (C)
	real, dimension(maxlines) :: Td		!Dew Point (C)
	character (len=1) :: tab=achar(9) 	!ascii tab character
end module soundmod

!============ main program ================
program skewtmain					!Skew-T plotting main program

!declare variables
	implicit none					!enforce strong typing

!set-up
	call welcome					!welcome the user
	call getfile					!read in the sounding file
	call prepareoutput				!prepare disk to receive output

!compute sounding plot data
	call plotsounding				!plot the sounding
	call plotisobar					!plot the isobars in the background
	call plotisotherm				!plot a few isotherms in the background
	call plotdryadiabat				!plot a few dry adiabats in background
	call plotwetadiabat				!plot a few saturated adiabats in backgr
	call plotisohume				!plot a few isohumes in the background

!finish
	call cleanup					!close files and release memory
	
end program skewtmain



!===============================================
subroutine welcome				!Welcome the interactive user
	implicit none					!enforce strong typing
	write(*,*)
	write(*,*) "================================================="
	write(*,*) "Welcome to Skew-T ... a Sounding Plotting Program"
	write(*,*) "================================================="
	write(*,*)
	write(*,*) "   Based on Stull,R., 2018: Practical Meteorology"
	write(*,*) "   Chapters 1 - 6. Coded by R. Stull, UBC, Feb 2007. Updated Aug 2021."
	write(*,*)
	write(*,*) "Expects ascii 'Text List' sounding as input, as from U.Wyoming site:"
	write(*,*) "   http://weather.uwyo.edu/upperair/sounding.html"
	write(*,*) "Produces .csv output file, which can be read into"
	write(*,*) "   Excel as one long Series for plotting."
	write(*,*)
end subroutine welcome


!===============================================
subroutine getfile						!read in the sounding file
	use soundmod						!use sounding module
	implicit none						!enforce strong typing
	character (len=100) :: line			!holds one line from the sounding
	integer :: ios						!input/output error status
	integer :: i						!dummy iteration counter
	real :: PhPa, zm, TC, TdC			!temporary Pressure, height, Temp, Td variabl

	write(*,*) "...starting getfile"	!debugging output
		
!First, ask the user for the file name that holds the sounding ascii data
	do i = 1,3							!give user 3 tries to enter correct file
		write(*,*)
		write(*,"(a)",advance="no") "Type in file name for sounding, then hit Enter: "
		read(*,*) filename
		
!Next, open the file
		open(1,file=filename,status="old",iostat=ios)	!connect to sounding file
		if (ios .ne. 0) then							!Can't open the file
			write(*,*) "Sorry, can't find file: ",filename
			write(*,*) "Don't forget to include the suffix .txt in the file name."
			if (i < 3) write(*,*) "Try again."
			cycle										!allow user to try again
		else											!Can open the file
			write(*,*) "Good.  Successfully opened: ", filename
			write(*,*)
			exit										!jump out of do loop
		endif
	enddo
	if (ios .ne. 0)	stop "Sorry. Perhaps the file doesn't exist.  Bye."

!At this point, we have a good file.

!Read title line and skip other header lines in the data file
	do i = 1,5										!for all 5 header lines
		read(1,"(a)",iostat=ios) line				!ios is newly set here
		if (ios .ne. 0) exit						!Cannot read lines
		if (i .eq. 1) title = line					!save title
	enddo	
	
	write(*,*) title								!Display title
	write(*,*)
	write(*,*) " i      P (kPa)      z (m)       T (C)      Td (C)"

!Read each data line in the sounding file, and echo to screen
	i = 0											!initialize line counter
	do												!for each sounding level
		read(1,"(F7.1, F7.0, 2F7.1)",iostat=ios) PhPa, zm, TC, TdC	!read obs
		if (ios .ne. 0) exit						!finished reading lines
		if ((PhPa < 100.0) .or. (i >= maxlines)) exit	!don't care about low P.  
		
		!At this point, a good line has been found, so save in data arrays.
		i = i + 1									!count data lines
		P(i) = 0.1*PhPa								!store Pressure, convert kPa
		z(i) = zm									!store heights
		T(i) = TC									!store temperatures
		Td(i) = TdC									!store dew points
		write(*,"(I3,4F12.2)") i, P(i), z(i), T(i), Td(i)	!echo line to screen
	enddo
	nlines = i										!save line count
	write(*,*)
	write(*,*) "Number of sounding lines captured = ",nlines

!Finished getting the sounding data.  Close the open file.
	write(*,*)
	close(1)						!close the connection to the input file
end subroutine getfile


!================================================
subroutine prepareoutput						!prepare disk to receive output
	use soundmod								!include sounding variables
	implicit none								!enforce strong typing
	integer :: i								!position of "." in filename
	integer :: ios								!error status for file opening
	
	write(*,*) "...starting prepareoutput"		!debugging output
	write(*,*)

!Create an output file to hold the results
	i = index(filename,".")	- 1					!find the character just before "." in the filename
	if (i .le. 0) i = len_trim(filename)		!use whole filename if no "."
	outname = filename(1:i) // "out.csv"		!create a new output file name
	write(*,*) "Output will be saved in file: ",outname
	write(*,*)
	open(2,file=outname,status="replace",iostat=ios)	!open the output file
	if (ios .ne. 0) then						!error opening file
		write(*,*) "Sorry, can't create file: ", outname
		stop "Bye."
	endif
	
!Write the title (sounding time and place) from the first header line of the input file
	write(*,*) trim(title)						!display title on screen
	write(*,*)
	write(*,*) " Calculated in addition to the sounding temperature and dew point:"
	write(*,*) "   Isobars:  P (kPa) = 100, 90, 80, 70, ..., 20, 10"
	write(*,*) "   Isotherms:  T (C) = -80, -60, -40, -20, 0, 20"
	write(*,*) "   Dry Adiabats: Theta (C) = -20, 0, 20, 40, 60, 80"
	write(*,*) "   Saturated adiabat: ThetaL (C) = 0, 10, 20, 30"
	write(*,*) "   Isohumes: r (g/kg) = 0.2, 1.0, 5.0, 20.0"
	write(*,*)
	write(*,*) "    X(T)    ",tab,"     Y(P)"  	!column headers
	
	write(2,*) trim(title)						!display title on output file
	write(2,*)
	write(2,*) " Calculated in addition to the sounding temperature and dew point:"
	write(2,*) "   Isobars:  P (kPa) = 100  90  80  70 ...  20  10"
	write(2,*) "   Isotherms:  T (C) = -80  -60  -40  -20  0  20"
	write(2,*) "   Dry Adiabats: Theta (C) = -20  0  20  40  60  80"
	write(2,*) "   Saturated adiabat: ThetaL (C) = 0  10  20  30"
	write(2,*) "   Isohumes: r (g/kg) =  0.2  1.0  5.0  20.0"
	write(2,*)
	write(2,*) "    X(T)    ", "," ,"     Y(P)"  	!column headers
	
	write(*,*)
end subroutine prepareoutput


!================================================
subroutine plotsounding							!plot the sounding
	use soundmod								!include sounding arrays
	implicit none								!enforce strong typing
	integer :: i								!dummy index
	real :: xst, yst							!function names
	real :: y									!ordinate value
	real :: xT, xTd								!abscissa values for T and Td(C)
	character (len=40) :: fmt					!format for output
	
	write(*,*) "...starting plotsounding"		!debugging output
	write(*,*)
	write(2,*)									!skip line in output file

	fmt = "(F12.4, a1, F12.4)"					!format for output

!	For the temperature profile
	do i=1,nlines								!for each sounding level
		y = yst(P(i))							!find ordinate value
		xT = xst(T(i),y)
		if ((xT .ge. -40.) .and. (xT .le. 40.)) then	!inside skewT domain
			write(*,fmt) xT,tab,y 				!Y & x coord for T, show on screen
			write(2,fmt) xT,",",y 				!Y & x coord fpr T, write in output file
		endif
	enddo
	
	write(*,*)
	write(2,*)									!skip line in output file
	
!	For the dew point profile
	do i=1,nlines								!for each sounding level
		y = yst(P(i))							!find ordinate value
		xTd = xst(Td(i),y)
		if ((xTd .ge. -40.) .and. (xTd .le. 40.)) then	!inside skewT domain
			write(*,fmt) xTd,tab,y 				!Y & x coord for Td, show on screen
			write(2,fmt) xTd,",",y 				!Y & x coord for Td, write in output file
		endif
	enddo
	
	write(*,*)
	write(2,*)									!skip line in output file
end subroutine plotsounding


!================================================
subroutine plotisobar							!plot several isobars in background
!Because isobars are straight lines, we need to find only the left and right ends
	use soundmod
	implicit none								!enforce strong typing
	integer :: i								!dummy height index
	real :: yst									!for the function subroutine
	real, parameter :: xcold = -40.0			!degC cold (left) end of isobar
	real, parameter :: xhot = 40.0				!degC hot (right) end of isobar
	real :: PkPa								!pressure (kPa)
	real :: y									!Y coordinate on skewT
	character(len=30) :: fmt					!string holding output format
	
	write(*,*) "...starting plotisobar"			!debugging output
	write(*,*)
	write(*,*) "Isobars:"
	write(*,*) "    X(T)    ",tab,"     Y(P)"	!column headers for output
	write(*,*)
	
	fmt = "(F12.4, a1, F12.4)"					!format for output
	write(2,*)									!skip line in output file
	write(2,*)									!skip line in output file

	do i = 10, 100, 10							!for each isobar to be plotted
		PkPa = real(i)							!convert pressure to real number
		y = yst(PkPa)							!get Y coordinate on skewT
		
		write(*,"(a,F10.1)") "P (kPa) = ",PkPa	!debug output on screen
		
		write(*,fmt) xcold,tab,y				!write isobar table to screen
		write(*,fmt) xhot,tab,y
		write(*,*)
		
		write(2,fmt) xcold,",",y				!write isobar table to output
		write(2,fmt) xhot,",",y
		write(2,*)
	enddo
		
	write(*,*)
	write(2,*)									!skip line in output file
end subroutine plotisobar


!================================================
subroutine plotisotherm							!plot a few isotherms in the background
	use soundmod								!include sounding data
	implicit none								!enforce strong typing
	real :: xst, yst							!type declaration for functions
	real :: x, y								!temporary coordinates in skewT
	real :: TC									!temperature (C)
	real :: PkPa								!pressure (kPa)
	integer :: i								!index for pressure height
	integer :: j								!index for temperature
	character(len=30) :: fmt					!string holding output format

	write(*,*) "...starting plotisotherm"		!debugging output
	write(*,*)
	write(*,*) "Isotherms:"
	write(*,*) "    X(T)    ",tab,"     Y(P)"	!column headers for output

	fmt = "(F12.4, a1, F12.4)"					!format for output

	do j = -80, 20, 20							!for each isotherm
		TC = real(j)							!temperature (C)
		write(*,*)
		write(*,"(a,F8.1)") "T (C) = ", TC
		write(2,*)
		
		do i = 100, 10, -5						!for each pressure height
			PkPa = real(i)						!pressure (kPa)
			y = yst(PkPa)						!y coordiinate of skewT
			x = xst(TC,y)						!x coordinate of skewT

			write(*,fmt) x,tab,y 				!x & y coord, show on screen
			if ((x .ge. -40.) .and. (x .le. 40.)) then	!inside skewT domain
				write(2,fmt) x,",",y 			!x & y coord, write to output
			endif
		enddo
	enddo										!end of isotherm loop
	
	write(*,*)
end subroutine plotisotherm


!================================================
subroutine plotdryadiabat						!plot a few dry adiabats in background
	use soundmod								!include sounding data
	implicit none								!enforce strong typing
	real :: xst, yst							!type declaration for functions
	real :: x, y								!temporary coordinates in skewT
	real :: TC, TK								!temperature (C) & (K)
	real :: thetaC, thetaK						!potential temperature (C) & (K)
	real :: PkPa								!pressure (kPa)
	real, parameter :: RdCp = 0.28571			!Rd/Cp in eq for pot temperature
	real, parameter :: Po = 100.				!reference pressure (kPa)
	integer :: i								!index for pressure height
	integer :: j								!index for potential temperature
	character(len=30) :: fmt					!string holding output format

	write(*,*) "...starting plotdryadiabat"		!debugging output
	write(*,*)
	write(*,*) "Dry Adiabats:"

	fmt = "(F12.4, a1, F12.4)"					!format for output
	write(2,*)
	write(2,*)

	do j = -20, 80, 20							!for each adiabat
		thetaC = real(j)						!potential temperature (C)
		thetaK = thetaC + 273.					!potential temperature (K)
		write(*,*)
		write(*,"(a,F8.1)") "Theta (C) = ", thetaC
		write(*,*) "    P (kPa)   ", tab, "    T (C)"
		write(2,*)
		
		do i = 100, 10, -5						!for each pressure height
			PkPa = real(i)						!pressure (kPa)
			TK = thetaK*( (PkPa/Po)**RdCp )		!temperature (K)
			TC = TK - 273.						!temperature (C)
			write(*,fmt) PkPa, tab, TC 			!debugging output			
			
			y = yst(PkPa)						!y coordiinate of skewT
			x = xst(TC,y)						!x coordinate of skewT

			if ((x .ge. -40.) .and. (x .le. 40.)) then	!inside skewT domain
				write(2,fmt) x,",",y 			!x & y coord, write to output
			endif
			
		enddo									!end of pressure height loop
	enddo										!end of adiabat loop
	
	write(2,*)
	write(*,*)
end subroutine plotdryadiabat


!================================================
subroutine plotwetadiabat						!plot a few saturated adiabats in backgr
	use soundmod								!include sounding data
	implicit none								!enforce strong typing
	integer :: j								!thetaL index
	integer :: i								!pressure (height) index
	real, parameter :: dP = 0.10000				!pressure increment delta P (kPa)
	real, parameter :: a = 0.28571				!dimensionless const for sat adiabat
	real, parameter :: b = 1.35E7				!(K^2) constant for sat adiabat
	real, parameter :: c = 2488.4				!(K) constant for saturated adiabat
	real, parameter :: LR = 5423.				!Lv/Rv (K) in Clausius-Clapeyon eq
	real, parameter :: eo = 0.611				!(kPa) reference vapor pressure
	real, parameter :: Toinv = 1./273.			!inverse of T (K) const in Clausius Clap eq
	real, parameter	:: eps = 0.622				!(g/g) epsilon = ratio of gas constants
	real :: TC, TK								!temperature of rising air parcel (C) and (K)
	real :: es									!(kPa) saturated vapor pressure
	real :: rs									!(g/g) saturated mixing ratio
	real :: PkPa								!pressure (kPa)
	real :: dTdP								!(K/kPa) change of T with P along sat adiabat
	real :: num, den							!temporary numerator & denominator of dT/dP eq.
	real :: xst, yst							!type declaration for functions
	real :: x, y								!temporary coordinates in skewT
	character(len=30) :: fmt					!string holding output format
	
	write(*,*) "...starting plotwetadiabat"		!debugging output
	write(*,*)
	write(2,*)									!skip line in output file
	write(2,*)									!skip line in output file

	fmt = "(F12.4, a1, F12.4)"					!format for output

	do j = 0, 30, 10							!for each thetaL
		TC = real(j)							!initialize thetaL (C)
		PkPa = 100.								!initialize Pressure (kPa)
		write(*,"(a,F10.1)") "Theta-L (C) = ",TC	!debugging output
		write(*,*) "    P(kPa)  ",tab,"     T(C)" 	!debugging output
	
		do i = 1, 1000							!for each pressure height
			if (modulo(i,50) .eq. 1) then		!for every 50th data point
				write(*,"(F10.1,a1,F10.2)") PkPa, tab, TC	!display data along sat adiabat
				y = yst(PkPa)					!Y coordinate on skewT
				x = xst(TC,y)					!x coordinate on skewT
				if ((x .ge. -40.) .and. (x .le. 40.)) then	!inside skewT domain
					write(2,fmt) x,",",y 		!x & y coord for Td, write to output file
				endif
			endif

			TK = TC + 273.						!thetaL (K)
			es = eo*exp(LR*(Toinv - (1/TK)))	!(kPa) sat vap press from Clausius-Clap
			rs = (eps*es) / (PkPa - es)			!(g/g) sat mixing ratio
			num = (a*TK) + (c*rs) 				!numerator of dT/dP eq
			den = PkPa * (1 + (b*rs/(TK*TK)))	!denominator of dT/dP eq
			dTdP = num/den						!temp gradient of sat adiabat
			
			TC = TC - dTdP*dP					!new temperature along sat adiabat
			PkPa = PkPa - dP					!at new pressure
			if (PkPa .le. 10.) exit				!reached top of diagram
		enddo									!end of pressure height loop
	
		write(2,*)
		write(*,*)
	enddo										!end of thetaL loop
	
end subroutine plotwetadiabat


!================================================
subroutine plotisohume							!plot a few isohumes in the background
!Uses Clausius-Clapeyron (eq 3 on crib sheet), for saturation with respect to liquid water
	use soundmod								!include sounding data
	implicit none								!enforce strong typing
	integer :: i								!index for pressure (height)
	integer :: j								!index for mixing ratio
	real, dimension(4), parameter :: r = (/ 0.2 ,1. ,5. ,20. /)	!mixing ratios (g/kg)
	real, dimension(5), parameter :: PkPa = (/100., 70., 40., 20., 10. /)	!pressure (kPa)
	real, parameter :: RvLv = 0.0001844			!Rv/Lv, in units of 1/K
	real, parameter :: Toinv = 1./273.			!inverse of T (K) const in Clausius Clap eq
	real, parameter	:: eps = 0.622				!(g/g) epsilon = ratio of gas constants
	real, parameter :: eo = 0.611				!(kPa) reference vapor pressure
	real :: lnarg								!argument of ln in eq (3)
	real :: sbarg								!argument inside square bracket in (3)
	real :: TdC									!dew point (C) temporary variable
	real :: rg									!mixing ratio (g/g) temporary variable
	real :: x, y								!X and Y coordinates on skew T
	real :: xst , yst							!type declaration for functions
	character(len=30) :: fmt					!string holding output format
	
	
	write(*,*) "...starting plotisohume"		!debugging output on screen
	write(*,*)
	write(*,*) "Isohumes:"

	fmt = "(F12.4, a1, F12.4)"					!format for output
	write(2,*)									!skip line in output file
	write(2,*)									!skip line in output file
	
	do j=1,4									!for each humidity
		write(*,*)								!debugging output
		write(*,"(a,F8.2)") "r (g/kg) = ",r(j)			!debugging output
		write(*,*) "    P(kPa)  ",tab,"     Td(C)" 	!debugging output
		rg = 0.001 * r(j)						!convert mixing ratio to (g/g)
		
		do i = 1,5								!for each pressure
			lnarg = (rg*PkPa(i)) / (eo*(rg+eps))	!argument of ln in eq (3)
			sbarg = (Toinv - (RvLv*log(lnarg)))		!argument in square brackets in (3)
			TdC = (1./sbarg) - 273.				!dew point (C), from eq (3)
			write(*,fmt) PkPa(i), tab, TdC				!debugging output
			
			y = yst(PkPa(i))					!Y coordinate on skew T
			x = xst(TdC,y)						!X coordinate on skew T
			
			if (x .le. 40.) write(2,fmt) x,",",y	!write coord if inside skewT domain
		enddo
		
		write(2,*)								!blank line between each isohume data	
	enddo
	
	write(*,*)
end subroutine plotisohume


!================================================
subroutine cleanup								!close files and release memory
	use soundmod								!including sounding info
	implicit none								!enforce strong typing
	write(*,*) "...starting cleanup"			!debugging output
	write(*,*)
	
	write(*,*) title							!remind user of title
	write(*,*) "Your sounding output file is: ",outname
	close(2)									!close the output file
	
	write(*,*)
	write(*,*) "Bye."
	write(*,*)
end subroutine cleanup


!================================================
real function yst (P)					!finds ordinate (y) value on skewT
	implicit none						!enforce strong typing
	real, intent(in) :: P				!pressure (kPa)
	real, parameter :: Po = 100.0		!reference pressure (kPa)
	yst = log(Po/P)						!ordinate for skewT - logP
end function yst


!================================================
real function xst (T,y)					!finds abscissa (x) value on skewT
	implicit none						!enforce strong typing
	real, intent(in) :: T				!temperature (C)
	real, intent(in) :: y				!ordinate (dimensionless)
	real, parameter :: K = 35.0			!reference temperature (C)
	xst = T + K*y						!abscissa for skewT - logP
end function xst


!======= end ==
		
